import os
import inspect

l_= lambda *x, sep=' ': sep.join(str(x) for x in x if x)
b_ = lambda *x, sep='\n': l_(*x, sep=sep)
fw_ = lambda *x, w=20, right=False: f"{l_(*x):{'>' if right else '<'}{w}}"
bb_ = lambda *x, ch='-', n=20: b_(ch*n, *x, ch*n)

def tbl_(*rows, buffer=2):
    rows = [[str(c) for c in cs] for cs in rows]
    widths = [max((len(c) for c in cs), default=0) for cs in zip(*rows)]
    return b_(*(l_(*(fw_(c, w=w + buffer-1) for c, w in zip(cs, widths))) for cs in rows))

def _filter_stack(stack, match=None, stop=None, offset=0, limit=None):
    if isinstance(stop, str):
        stopstr, stop = stop, lambda f: stopstr == f.function
    elif stop is None:
        stop = lambda f: False
    else:
        stopframe, stop = stop, lambda f: stopframe == f
    if isinstance(match, str):
        includestr, match = match, lambda f: (
            includestr == f.function or
            includestr in f.filename)

    x = []
    for i, f in enumerate(stack[offset+1:][:limit]):
        if match and not match(f): continue
        if stop(f): break
        x.append(f)
    return x

def format_stack(message=None, match=None, stop=None, offset=0, limit=None):
    '''Format the current stack trace.'''
    return bb_(message, tbl_(*(
        (f.function, f.lineno, f.filename, f'>>> {f.code_context[0].strip()}')
        if i else 
        (f.function, f.lineno, f.filename, '')
        for i, f in enumerate(_filter_stack(
            inspect.stack(), match, stop, offset=offset, limit=limit))
    )))

def print_stack(*a, offset=0, **kw):
    '''Print out the current stack trace.'''
    print(format_stack(*a, offset=offset + 1, **kw), flush=True)


def short_stack(match=None, file=False, sep=' << ', format='{}', stop=None, offset=0, n=None):
    '''get a compressed view of the stack.'''
    return sep.join(
        format.format(f'{f.function} ({os.path.basename(f.filename)}:{f.lineno})' if file else f.function)
        for f in _filter_stack(inspect.stack(), match, stop, offset+1, n)
    )

def funcstr(func, *a, **kw):
    return '{}({})'.format(func.__name__, ', '.join(
        ['{!r}'.format(v) for v in a] +
        ['{}={!r}'.format(k, v) for k, v in kw.items()]
    ))

def excline(e):
    return '{}: {}'.format(type(e).__name__, e)






if __name__ == '__main__':
    def aaa(**kw):
        bbb(**kw)
    def bbb(**kw):
        ccc(**kw)
    def ccc(**kw):
        print_stack(**kw)

    def nested(func, *a, N=3, **kw):
        if N>0:
            return nested(func, *a, N=N-1, **kw)
        func(*a, **kw)

    def main():
        aaa()
        aaa(match='debug')
        aaa(stop=inspect.stack()[0])
        nested(lambda: print(short_stack()))
        nested(lambda: print(short_stack(match='debug.py')))
        nested(lambda: print(short_stack(stop='main')))
    import fire
    fire.Fire(main)


import ast

def parse(value):
    """Parse a python literal from a string, allowing for some flexibilities in syntax.
    
    .. code-block:: python
    
        import starstar.parse.parse as ssparse
        assert ssparse('5') == 5
        assert ssparse('asdf') == 'asdf'
        assert ssparse('True') == True
        assert ssparse('None') == None
        assert ssparse('[1,2,3]') == [1, 2, 3]
        assert ssparse('[1,2,asdf]') == [1, 2, 'asdf']
        assert ssparse('{a:5}') == {'a': 5}
        assert ssparse('{a:[1,2,3]}') == {'a': [1, 2, 3]}
    """
    try:
        return _literal_eval(value)
    except (SyntaxError, ValueError):
        return value


def _literal_eval(value):
    root = ast.parse(value, mode='eval')
    if isinstance(root.body, ast.BinOp):  # pytype: disable=attribute-error
        raise ValueError(value)

    for node in ast.walk(root):
        for field, child in ast.iter_fields(node):
            if isinstance(child, list):
                for index, subchild in enumerate(child):
                    if isinstance(subchild, ast.Name):
                        child[index] = _replacement(subchild)
            elif isinstance(child, ast.Name):
                node.__setattr__(field, _replacement(child))

    # supports: strings, bytes, numbers, tuples, lists, dicts, sets, booleans, and None
    return ast.literal_eval(root)


_BUILTIN = ('True', 'False', 'None')  # TODO: '...'
def _replacement(node):
  """Returns a node to use in place of the supplied node in the AST."""
  value = node.id
  return node if value in _BUILTIN else ast.Str(value)


if __name__ == '__main__':
    def testit(s, v):
        print(f'input: {s!r}')
        print(f'expected: {v!r}')
        print(f'got: {parse(s)!r}')
        print()

    def main():
        testit('5', 5)
        testit('asdf',  'asdf')
        testit('True',  True)
        testit('None',  None)
        testit('[1,2,3]',  [1, 2, 3])
        testit('[1,2,asdf]',  [1, 2, 'asdf'])
        testit('{a:5}',  {'a': 5})
        testit('{a:[1,2,3]}',  {'a': [1, 2, 3]})
    import fire
    fire.Fire(main)


from __future__ import annotations
import inspect
from types import MappingProxyType
from typing import Callable, Iterable, cast as tcast
from inspect import signature as _builtin_signature, Signature as _Signature
from functools import wraps as _builtin_wraps, update_wrapper as _update_wrapper


POS_ONLY = inspect.Parameter.POSITIONAL_ONLY
KW_ONLY = inspect.Parameter.KEYWORD_ONLY
POS_KW = inspect.Parameter.POSITIONAL_OR_KEYWORD
VAR_POS = inspect.Parameter.VAR_POSITIONAL
VAR_KW = inspect.Parameter.VAR_KEYWORD

ALL = {POS_ONLY, KW_ONLY, POS_KW, VAR_POS, VAR_KW}
POS = {POS_ONLY, POS_KW}
KW = {POS_KW, KW_ONLY}
VAR = {VAR_POS, VAR_KW}

NAMED = ALL - VAR
NOT_KW = ALL - KW



def divide(kw: dict, *funcs: Callable|Iterable[Callable], mode='strict', varkw: bool=True):
    '''Divide ``**kwargs`` between multiple functions based on their signatures.
    
    Arguments:
        *funcs (callable): The functions you want to divide arguments amongst.
        mode (str): How to handle extra keyword arguments.

             - ``'separate'``: Add them as an extra dictionary at the end.
             - | ``'strict'``: If there are extra keyword arguments, raise a ``TypeError``. 
                This will only raise if no function with a variable keyword is found or ``varkw == False``.
        varkw (bool | 'first'): Do we want to pass extra arguments as ``**kwargs`` to any function 
            that accepts them? By default (True), they will go to all functions with variable keyword 
            arguments. If ``'first'`` they will only go to the first matching function. If ``False``, 
            no function will get variable keyword args.

    Returns:
        A dict for each function provided, plus one more for unused kwargs if ``mode == 'separate'``.

    Raises:
        TypeError: if ``mode='strict'`` (default) and it receives arguments that don't appear in any function's signature. Does not apply if any function takes ``**kw``.

    Pass arguments to multiple functions!

    .. code-block:: python 

        def func_a(a=None, b=None, c=None):
            return a, b, c

        def func_b(d=None, e=None, f=None):
            return d, e, f

        def main(**kw):
            kw_a, kw_b = starstar.divide(kw, func_a, func_b)
            func_a(**kw_a)
            func_b(**kw_b)

        # and it even works for nested functions !

        def func_c(**kw):
            kw_a, kw_b = starstar.divide(kw, func_a, func_b)
            func_a(**kw_a)
            func_b(**kw_b)

        def func_d(g=None, h=None, i=None):
            return g, h, i

        def main(**kw):
            kw_c, kw_d = starstar.divide(kw, (func_a, func_b), func_d)
            func_c(**kw_c)
            func_d(**kw_d)

    Decide how you want to handle extra keyword args.

    .. code-block:: python

        main(a=1, x=2)  # extra argument "x"
        # divide raises TypeError

        def main(**kw):
            kw_a, kw_b, kw_extra = starstar.divide(kw, func_a, func_b, mode='separate')

        main(a=1, x=2)  # extra argument "x"
        # gets put in ``kw_extra``
    '''
    pss = [_nested_cached_sig_params(f) for f in funcs]
    kws = [{} for _ in pss]

    # get all keys that match explicitly defined parameters
    kwunused = dict(kw)
    for ps, kwi in zip(pss, kws):
        for name in ps:
            if ps[name].kind in NOT_KW or name not in kw:
                continue
            kwi[name] = kw[name]
            kwunused.pop(name, None)

    # check functions for varkw
    if kwunused and varkw:
        found_varkw = False
        for ps, kwi in zip(pss, kws):
            if any(p.kind == VAR_KW for p in ps.values()):
                kwi.update(kwunused)
                found_varkw = True
                if varkw == 'first':
                    break
        if found_varkw:
            kwunused.clear()

    # handle extra kwargs
    if mode == 'separate':
        kws.append(kwunused)
    elif mode == 'strict':
        if kwunused:
            raise TypeError('Got unexpected arguments: {}'.format(tuple(kwunused)))

    return kws


def _nested_cached_sig_params(f: Callable|list|tuple) -> dict|MappingProxyType:
    '''Get and merge signature parameters for potentially multiple functions.'''
    # return signature(f).parameters
    # return {k: p for fi in _nested(f) for k, p in signature(fi).parameters.items()}
    if isinstance(f, (list, tuple)):
        return {k: v for fi in f for k, v in signature(fi).parameters.items()}
    return signature(f).parameters

def _nested(xs, types=(tuple, list)):
    if isinstance(xs, types):
        yield from (xi for x in xs for xi in _nested(x))
    else:
        yield xs


def signature(f: Callable, required=True) -> _Signature:  # type: ignore
    '''Get a function signature.
    
    Faster than inspect.signature (after the first call) because it 
    is cached using the standard ``f.__signature__`` attribute.
    '''
    try:
        try:
            return f.__signature__
        except AttributeError:
            s = tcast(_Signature, _builtin_signature(f))
            try:
                f.__signature__ = s
            except AttributeError:
                try:
                    f.__dict__['__signature__'] = s
                except AttributeError:
                    pass
            return s
    except ValueError:
        if required:
            raise

def traceto(*funcs: Callable, keep_varkw=None, filter_hidden=True, doc=False) -> Callable:  # , kw_only=True
    '''Tell a function where its ``**kwargs`` are going!

    This is similar to ``functools.wraps``, except that it merges the signatures of multiple functions
    and only deals with arguments that can be passed as keyword arguments.

    Arguments:
        *funcs (callable): The functions where the keyword arguments go.
        keep_varkw (bool): Whether we should keep ``**kw`` in the signature.
            If not set, this will be True if any of the passed arguments take 
            variable kwargs.
        filter_hidden (bool): Whether we should filter out arguments starting with 
            an underscore. Default True. This is often used for private arguments
            for example in the case of recursive functions that pass objects internally.
        doc (bool): Whether to merge the docstrings. Defaults to False.

    By having ``func_c`` trace its ``**kwargs`` signature, ``main`` can say that it's passing
    its arguments to ``func_c`` and it will be able to see the parameters of ``func_a`` & ``func_b``.

    .. code-block:: python

        @starstar.traceto(func_a, func_b)
        def func_c(**kw):
            kw_a, kw_b = starstar.divide(kw, func_a, func_b)
            func_a(**kw_a)
            func_b(**kw_b)

        @starstar.traceto(func_c, func_d)
        def main(**kw):
            kw_c, kw_d = starstar.divide(kw, func_c, func_d)
            func_c(**kw_c)
            func_d(**kw_d)

    ..
        .. note::
            I hope to have a way of tracing positional arguments to a single function as well. But it adds 
            more complexity to keyword dividing and I don't think it's worth it until we can flesh out 
            nice, organized, clearly defined, and intuitive behavior.

            e.g. what to do with this?

            .. code-block:: python

                def func_a(a, b, c): pass
                def func_b(b, c, d): pass
                def func_c(*a, **kw):
                    starstar.divide(a, kw, func_a, func_b)
                func_c(1, 2)  # should func_b get b=2 ???
                # if not, how can I pass b=2 to func_b ?
    '''
    # get parameters from source functions
    f_params = [_nested_cached_sig_params(f) for f in funcs]
    ps_all = [p for ps in f_params for p in ps.values()]
    if keep_varkw is None:  # check if any have varkw
        keep_varkw = any(p.kind == VAR_KW for p in ps_all)

    # remove private parameters (start with '_')
    if filter_hidden:
        ps_all = [p for p in ps_all if not p.name.startswith('_')]
    # remove duplicates
    ps = {p.name: p for p in ps_all if p.kind not in NOT_KW}
    # make the parameters kwonly
    other_ps = [p.replace(kind=KW_ONLY) for p in ps.values()]# if kw_only else list(ps.values())

    def decorator(func):
        # copy func
        @_builtin_wraps(func)
        def f(*a, **kw):
            return func(*a, **kw)

        # get signature from the decorated function
        sig = signature(f)
        params = tuple(sig.parameters.values())
        p_poskw = [p for p in params if p.kind in (POS_ONLY, POS_KW, VAR_POS, KW_ONLY)]
        p_varkw = [p for p in params if p.kind == VAR_KW]
        p_names = {p.name for p in params if p.kind not in VAR}
        other_ps_ = [p for p in other_ps if p.name not in p_names]

        if doc:
            f.__doc__ = _mergedoc(f.__doc__, funcs, other_ps_) #MergedDocstring(f.__doc__, funcs, other_ps_)

        # merge parameters and replace the signature
        f.__signature__ = sig.replace(parameters=(
            p_poskw + other_ps_ + (p_varkw if keep_varkw else [])))
        f.__starstar_traceto__ = funcs
        return f
    return decorator



def wraps(func: Callable, skip_args=(), skip_n=0) -> Callable:
    '''``functools.wraps``, except that it merges the signature.

    .. note::

        ``functools.wraps`` doesn't do any function introspection. This means
        that if the wrapper function adds any arguments to the function 
        signature, these arguments won't be documented.
    
    If you're not familiar with ``functools.wraps``, it is a decorator
    that renames a wrapper function and its signature to look like the 
    function that it's wrapping.

    .. code-block:: python

        # without wrapping

        def print_output(func):
            def inner(*a, print_output=True, **kw):
                output = func(*a, **kw)
                if print_output:
                    print(output)
                return output
            return inner

        @print_output
        def something(a, b):
            return a+b

        assert something.__name__ == 'inner'

        # with wrapping

        def print_output(func):
            @starstar.wraps(func)
            def inner(*a, print_output=True, **kw):
                output = func(*a, **kw)
                if print_output:
                    print(output)
                return output
            return inner

        @print_output
        def something(a, b):
            return a+b

        assert something.__name__ == 'something'
    '''
    wrap = _builtin_wraps(func)
    def decorator(wrapper):
        sig = _merge_signature(wrapper, func, skip_args, skip_n)
        f = wrap(wrapper)
        f.__signature__ = sig
        return f
    return decorator
# wraps.__doc__ += '\n\n\n' + _builtin_wraps.__doc__


def _merge_signature(wrapper, wrapped, skip_args=(), skip_n=0):
    '''Merge the signatures of a wrapper and its wrapped function.'''
    sig = signature(wrapped)
    sig_wrap = signature(wrapper)
    psposkw, psvarpos, pskw, psvarkw = _param_groups(sig, skip_args)
    pswposkw, _, pswkw, _ = _param_groups(sig_wrap)
    return sig.replace(parameters=(
        pswposkw + psposkw[skip_n or 0:] + psvarpos + pswkw + pskw + psvarkw))

def _param_groups(sig, skip_args=()):
    '''Return the parameters by their kind. Useful for interleaving.'''
    params = tuple(sig.parameters.values())
    if skip_args:
        skip_args = asitems(skip_args)
        params = [p for p in params if p.name not in skip_args]
    return (
        [p for p in params if p.kind in (POS_ONLY, POS_KW)],
        [p for p in params if p.kind == VAR_POS],
        [p for p in params if p.kind == KW_ONLY],
        [p for p in params if p.kind == VAR_KW])


def partial(__func, *a_def, **kw_def):
    '''``functools.partial``, except that it updates the signature and wrapper.
    
    known bug: kw defaults dont update in signature.
    '''
    @wraps(__func, skip_n=len(a_def))
    def inner(*a, **kw):
        return __func(*a_def, *a, **{**kw_def, **kw})
    return inner

# given signature and dict, produce *a, **kw

def as_args_kwargs(func, kw):
    '''Separate out positional and keyword arguments using a function's signature.

    Sometimes functions have position only arguments, but you still want to be able 
    to configure them in a single dictionary. In order to pass them to a function
    you need to separate them out first. This will separate out all arguments that 
    can be passed as positional arguments.
    
    Arguments:
        func (callable): The function that the arguments will be passed to.
        values (dict): The parameter values you want to pass to the function.
            NOTE: This value is not modified.

    Returns: 
        (tuple): A tuple containing
         - a (list): The positional arguments we could pull out.
         - kw (dict): The keyword args that are left over.

    .. code-block:: python

        def func_a(a, b, c, *, d=0):
            return a, b, c, d

        # split out the args and kwargs
        a, kw = starstar.as_args_kwargs(func_a, {'a': 1, 'b': 2, 'c': 3, 'd': 4})
        assert a == [1, 2, 3]
        assert kw == {'d': 4}

        # 
        assert func_a(*a, **kw) == (1, 2, 3, 4)
        
    '''
    sig = signature(func)
    pos = []
    kw = dict(kw)
    for name, arg in sig.parameters.items():
        if arg.kind == VAR_POS:
            pos.extend(kw.pop(arg.name, ()))
            pos.extend(kw.pop('*', ()))
            break
        if arg.kind not in (POS_ONLY, POS_KW) or name not in kw:
            break
        pos.append(kw.pop(name))
    return pos, kw

# get arguments matching a condition

def get_args(func, match=(), ignore=()):
    '''Get argument parameters matching a specific type.

    Arguments:
        func (callable): The function to inspect.
        match (int or set): Argument types to filter for.
        ignore (int or set): Argument types to filter out.

    Returns:
        (list): a list containing the matching parameter objects.
    
    .. code-block:: python

        def func(x, y=3, *a, z=4, **kw):
            pass

        # get all arguments
        args = starstar.get_args(func)
        assert all(isinstance(p, inspect.Parameter) for p in args)
        assert [p.name for p in args] == ['x', 'y', 'a', 'z', 'kw']

        # get positional arguments
        args = starstar.get_args(func, starstar.POS)
        assert [p.name for p in args] == ['x', 'y']

        # get arguments excluding *a, **kw
        args = starstar.get_args(func, ignore=starstar.VAR)
        assert [p.name for p in args] == ['x', 'y', 'z']

        # get keyword arguments
        args = starstar.get_args(func, ignore=starstar.KW)
        assert [p.name for p in args] == ['x', 'y', 'z']

        # get keyword only arguments
        args = starstar.get_args(func, ignore=starstar.KW_ONLY)
        assert [p.name for p in args] == ['z']
    '''
    match = (set(asitems(match)) or ALL) - set(asitems(ignore))
    return [p for p in signature(func).parameters.values() if p.kind in match]

# get required arguments

def required_args(func):
    '''Get the required arguments for a function.'''
    return [
        p for p in signature(func).parameters.values() 
        if p.default is inspect._empty and p.kind not in VAR
    ]

# filter kw dict using function signature

def filter_kw(func, kw, skip_n=0, pop=False, inverse=False, unmatched=False, include_varkw=True):
    '''Filter ``**kwargs`` down to only those that appear in the function signature.
    
    Arguments:
        func (callable): The function to filter using.
        kw: keyword arguments that you'd like to filter.
        pop (bool): Remove matched keys from kw.
        inverse (bool): Return keys not in function signature.
        include_varkw (bool): if a function takes **kw, should it 
            swallow all arguments? default True.

    Returns:
        (dict): the filtered ``kwargs``

    .. code-block:: python

        def func_a(a, b, c): 
            pass

        args = {'b': 2, 'c': 3, 'x': 1, 'y': 2}
        assert starstar.filter_kw(func_a, args) == {'b': 2, 'c': 3}
    '''
    ps = list(signature(func).parameters.values())[skip_n or 0:]
    varkw = include_varkw and next((True for p in ps if p.kind == VAR_KW), False)
    ks = set(kw)
    ps = {p.name for p in ps if p.kind in KW}
    if varkw:
        ps = ps|ks
    if unmatched:
        return ks - ps if inverse else ps - ks
    ks = ks - ps if inverse else ks & ps
    return {k: kw.pop(k) for k in ks} if pop else {k: kw[k] for k in ks}


def filtered(func):
    '''A decorator that filters out any extra kwargs passed to a function.
    See ``starstar.filter`` for more information.

    .. code-block:: python

        @starstar.filtered
        def func_a(a, b, c): pass

        func_a(1, 2, c=3, x=1, y=2)  # just gonna ignore x and y

        # using the decorator is equivalent to
        func_a(*a, **starstar.filter_kw(func_a, kw))
    '''
    @_builtin_wraps(func)
    def inner(*a, **kw):
        return func(*a, **filter_kw(func, kw))
    return inner


def kw2id(kw, *keys, key=True, sep='-', key_sep='_', filter=True, missing='', format=str):
    '''Create an id from keyword arguments.
    
    Arguments:
        kw (dict): The available arguments.
        *keys (str): The keys to use.
        key (bool): Include the key in the id?
        sep (str): The separator between each item.
        key_sep (str): The separator between key and value (if ``key=True``).
        filter (bool): Filter out missing values.
        format (callable): A function that formats the values.

    .. code-block:: python

        kw = {'name': 'asdf', 'count': 10, 'enabled': True}
        assert starstar.kw2id(kw, 'name', 'count', 'enabled', 'xxx') == 'name_asdf-count_10-enabled_True'
        assert starstar.kw2id(kw, 'name', 'xxx', 'count', filter=False) == 'name_asdf-xxx_-count_10'
    '''
    ki = ((k, kw.get(k, missing)) for k in keys if not filter or k in kw)
    return sep.join(f'{k}{key_sep}{format(i)}' if key else f'{i}' for k, i in ki)



def asitems(x, types=(list, tuple, set)):
    '''Convert a value into a list/tuple/set. Useful for arguments that can be ``None, single item, list, tuple``.

    .. code-block:: python

        assert starstar.asitems(None) == []
        assert starstar.asitems('asdf') == ['asdf']
        assert starstar.asitems([1, 2, 3]) == [1, 2, 3]
    '''
    return x if isinstance(x, types) else (x,) if x is not None else ()



from .dcp_nesteddoc import _mergedoc #  circular


from __future__ import annotations
import inspect
from functools import update_wrapper as _update_wrapper
from .core import signature, POS_ONLY, POS_KW, VAR_KW, VAR


class _Defaults:
    def __init__(self, func, varkw=True, strict=False, frozen=None):
        self.function = func
        self.kw_defaults = {}
        self.__name__ = getattr(func, '__name__', None)

        _update_wrapper(self, func)
        sig = signature(func)
        ps = sig.parameters
        self.posargs = [k for k in ps if ps[k].kind in (POS_ONLY, POS_KW)]
        self.varkw = varkw and any(p.kind == VAR_KW for p in ps.values())
        self.strict = strict

        if func.__kwdefaults__ is None:
            func.__kwdefaults__ = {}
        self._backup = frozen or self._freeze()        

    def __str__(self):
        return '{}({}){}'.format(self.__class__.__name__, self.__name__, self.function.__signature__)

    def __call__(self, *a, **kw):
        pos = set(self.posargs[:len(a)])
        return self.function(*a, **{
            k: v for k, v in ({**self.kw_defaults, **kw}).items()
            if k not in pos
        })

    def update(self, **update):
        '''Update the function's defaults.'''
        update = update_defaults(self.function, **update)
        if update:
            if not self.varkw:
                raise TypeError('Unexpected arguments: {}'.format(set(update)))
            self.kw_defaults.update(update)
        return self

    def clear(self):
        '''Reset the function's defaults.'''
        self._restore(self._backup)
        return self

    def get(self):
        '''Get the function's defaults.'''
        return get_defaults(self.function)

    def _freeze(self):
        '''Get the current defaults/signature in case we want to restore back to this point.'''
        f = self.function
        return (signature(f), f.__defaults__, dict(f.__kwdefaults__), dict(self.kw_defaults))

    def _restore(self, frozen):
        '''Use the output of self._freeze() to restore to a previous set of defaults.'''
        f = self.function
        f.__signature__, f.__defaults__, f.__kwdefaults__, self.kw_defaults = frozen
        return self


def defaults(func, *a, **kw):
    '''Allow functions to have easily overrideable default arguments.
    Works with mixed positional and keyword arguments.

    This is a wrapper around ``update_defaults``, that also supports 
    default values for unnamed arguments (``**kwargs``).

    ``defaults(func).update`` will raise a ``TypeError`` if an extra 
    argument is passed, whereas ``update_defaults`` will return 
    the extra arguments.

    NOTE: This interface is not stable yet.

    .. code-block:: python 

        @starstar.defaults
        def abc(a=5, b=6):
            return a + b    

        assert abc() == 11
        abc.update(a=10)
        assert abc() == 16
        assert abc(2) == 8
    '''
    inner = lambda func: _Defaults(func, *a, **kw)
    return inner(func) if callable(func) else inner


def update_defaults(func, **update):
    '''Update a function's default arguments. Because functions don't 
    have a mechanism for defining default ``**kwargs``, this will return 
    any parameters not explicitly named in the signature. 

    TODO: should this be strict?

    .. code-block:: python 

        def abc(a=5, b=6):
            return a + b

        assert starstar.get_defaults(abc) == {'a': 5, 'b': 6}

        starstar.update_defaults(abc, b=7)

        assert starstar.get_defaults(abc) == {'a': 5, 'b': 7}
    '''
    sig = signature(func)
    ps = sig.parameters

    # update signature (do this before we pop)
    if set(ps) & set(update):
        func.__signature__ = sig.replace(parameters=[
            p.replace(default=update[k]) if k in update else p
            for k, p in ps.items()
        ])

    # update pos/poskw
    posargs = [k for k in ps if ps[k].kind in (POS_ONLY, POS_KW)]
    if func.__defaults__ and set(posargs) & set(update):
        func.__defaults__ = tuple(
            update.pop(name, current)
            for name, current in zip(posargs[::-1], func.__defaults__[::-1])
        )[::-1]

    # update kwonly/varkw
    kw_defaults = func.__kwdefaults__
    if kw_defaults and update:
        kw_defaults.update(
            (k, update.pop(k)) for k in set(update) if k in kw_defaults)

    # if not any(p.kind == VAR_KW for p in ps):  # 
    #     update.clear()
    return update


def get_defaults(func):
    '''Get the non-empty default arguments to a function (as a dict).
    
    .. code-block:: python 

        def abc(a=5, b=6):
            return a + b

        assert starstar.get_defaults(abc) == {'a': 5, 'b': 6}
    '''
    ps = ((k, p.default) for k, p in signature(func).parameters.items() if p.kind not in VAR)
    return {k: d for k, d in ps if d is not inspect._empty}



from __future__ import annotations
import inspect
from functools import wraps as _builtin_wraps
import docstring_parser as dcp
from .core import signature, _nested

# Needed for issue: https://github.com/rr-/docstring_parser/issues/61
def _build_meta(self, text, title):
    section = self.sections[title]

    if (
        section.type == dcp.google.SectionType.SINGULAR_OR_MULTIPLE
        and not dcp.google.MULTIPLE_PATTERN.match(text)
    ) or section.type == dcp.google.SectionType.SINGULAR:
        return self._build_single_meta(section, text)

    if ":" not in text:
        raise dcp.ParseError("Expected a colon in {}.".format(repr(text)))

    # Split spec and description
    before, desc = text.split(":", 1)
    if desc:
        desc = desc[1:] if desc[0] == " " else desc
        if "\n" in desc:
            lines = desc.splitlines(keepends=True)
            first_line, lines = lines[0], lines[1:]
            i = next((i for i, l in enumerate(lines) if l.strip()), 0)
            spaces = ''.join(lines[:i])
            rest = ''.join(lines[i:])
            desc = first_line + spaces + inspect.cleandoc(rest)
        desc = desc.strip("\n")

    return self._build_multi_meta(section, before, desc)
dcp.google.GoogleParser._build_meta = _build_meta



# class Docstring:
#     def __init__(self, doc, merge=True):
#         self.doc = doc or ''
#         self.merge = merge
#         self._parsed = None

#     def compile(self):
#         return self.doc

#     def __str__(self):
#         if not self.merge:
#             return self.doc or ''
#         if self._parsed is None:
#             self._parsed = self.compile()
#         return str(self._parsed or '')

# class MergedDocstring(Docstring):
#     def __init__(self, doc, funcs, ps=None, **kw):
#         super().__init__(doc, **kw)
#         self.funcs = funcs or []
#         self.ps = ps
        
#     def compile(self):
#         return mergedoc(self.doc, self.funcs, self.ps)

# class NestedDocstring(Docstring):
#     def __init__(self, doc, funcs, **kw):
#         super().__init__(doc, **kw)
#         self.funcs = funcs or {}

#     def compile(self):
#         return nestdoc(self.doc, self.funcs)


def _mergedoc(doc, funcs, ps=None, style=None):
    import docstring_parser as dcp
    funcs = list(_nested(funcs))  # unravel nested functions

    if ps is None:
        ps = [p for f in funcs for p in signature(f).parameters.values()]

    parsed = dcp.parse(doc)
    docstrs = [dcp.parse(f.__doc__ or '') for f in funcs]
    docps = [{p.arg_name: p for p in d.params} for d in docstrs]

    # find the style from the first function that isn't just a description
    style = style or next(
        (d.style for d in [parsed]+docstrs if d.meta), 
        parsed.style)

    # hack to add a new line between description and parameters
    # if the main function doesn't have params of its own.
    if not parsed.meta:
        parsed.blank_after_long_description = True

    # add params to main docstring
    params = []
    existing = {p.arg_name for p in parsed.params}
    for p in ps:
        if p.name in existing:
            continue
        for d in docps:
            if p.name in d:
                existing.add(p.name)
                params.append(d[p.name])
                break

    # find the right place to splice in the params
    meta = parsed.meta
    i = next((
        i+1 for i, m in list(enumerate(meta))[::-1] 
        if isinstance(m, dcp.DocstringParam)), 0)
    parsed.meta = meta[:i] + params + meta[i:]
    return dcp.compose(parsed, style=style)


def _nestdoc(doc, funcs, style=None):
    
    def _indented_doc_args(func):
        dd = dcp.parse(func.__doc__)
        dd.long_description = ''
        dd.short_description = ''
        dd.blank_after_long_description = False
        dd.blank_after_short_description = False
        dd.meta = [m for m in dd.meta if m.args[0] == 'param']
        for m in dd.meta:
            m.arg_name = f' - {m.arg_name}'
        return '\n'.join(dcp.compose(dd).lstrip().splitlines()[1:])

    parsed = dcp.parse(doc)
    docstrs = {k: dcp.parse(f.__doc__ or '') for k, f in funcs.items()}

    # find the style from the first function that isn't just a description
    style = style or next(
        (d.style for d in [parsed]+list(docstrs.values()) if d.meta), 
        parsed.style)

    # hack to add a new line between description and parameters
    # if the main function doesn't have params of its own.
    if not parsed.meta:
        parsed.blank_after_long_description = True

    # add params to main docstring
    params = []
    current_params = parsed.params
    existing = {p.arg_name for p in current_params}

    # add arguments
    for key, func in funcs.items():
        if key in existing:
            for p in current_params:
                if p.name == key:
                    desc = p.description or f'Keyword arguments for :func:`{func.__name__}`.'
                    p.description = f'{desc}\n\n{_indented_doc_args(func)}'
                    p.type_name = p.type_name or 'dict'
                    break
        else:
            params.append(dcp.DocstringParam(
                ['param', f'{key} (dict)'], 
                f'Keyword arguments for :func:`{func.__name__}`.\n\n{_indented_doc_args(func)}',
                arg_name=key, type_name='dict', 
                is_optional=True, default=None))

    # find the right place to splice in the params
    meta = parsed.meta
    i = next((
        i+1 for i, m in list(enumerate(meta))[::-1] 
        if isinstance(m, dcp.DocstringParam)), 0)
    parsed.meta = meta[:i] + params + meta[i:]
    return dcp.compose(parsed, style=style)


def nestdoc(*funcs, template='{}_kw', **named):
    '''Nest a function's docstring parameters as a dict in another
    function's parameters.
    
    Arguments:
        *funcs: functions to nest. The argument name is gotten using ``template``.
        template (str): The keyword argument template for ``*funcs``. 
            Defaults to ``'{}_kw'`` where {} is replaced with ``f.__name__``.
        **named: functions to nest. The key is used as the argument name
            in the parent function.

    .. code-block:: python

        def funcA(a, b):
            """Another function
            
            Arguments:
                a (int): a from funcA
                b (int): b from funcA
            """

        def funcB(b, c):
            """Anotherrrr function
            
            Arguments:
                b (int): b from funcB
                c (int): c from funcB
            """


        def funcC(**kw):
            """Hello"""
        funcD = nested_doc(funcA, funcB)(funcC)

        print(funcD.__doc__)
        """
        Hello

        Args:
            funcA_kw (dict?): Keyword arguments for ``funcA``.
                
                    - a (int): a from funcA
                    - b (int): b from funcA
            funcB_kw (dict?): Keyword arguments for ``funcB``.
                
                    - b (int): b from funcB
                    - c (int): c from funcB
        """

    '''
    allfuncs = {}
    for f in funcs:
        allfuncs[template.format(f.__name__)] = f
    allfuncs.update(named)
    if not all(callable(f) for f in named.values()):
        raise TypeError("all functions must be callable")

    def inner(func):
        @_builtin_wraps(func)
        def func2(*a, **kw):  # copies func
            return func(*a, **kw)
        func2.__doc__ = _nestdoc(func.__doc__, allfuncs) #str(NestedDocstring(func.__doc__, named))
        return func2
    return inner



# def funcA(a, b):
#     '''Another function
    
#     Arguments:
#         a (int): blah
#         b (int): blallhak
#     '''

# def funcB(b, c):
#     '''Anotherrrr function
    
#     Arguments:
#         b (int): blah
#         c (int): blallhak
#     '''


# def funcC(**kw):
#     '''Hello
    
#     Arguments:
#         x: asdfasdf
            
            
#             asdfasdf
#             asdf
#     '''
# funcD = nestdoc(funcA, funcB)(funcC)



"""

.. note::
    
    It currently works for Google and Numpy docstrings, but it's still a bit of a WIP.

It tries its best to preserve the surrounding whitespace, and it will separate out
whitespace, below, above, and to the left (common indentation) of the text block, so 
you can edit the content of the block and the rest of it will be preserved.

.. code-block:: python

    import starstar.docstr as ssdoc

    doc = ssdoc.parse('''This is my docstring description.

    Args:
        a (str): this is a
        b (str): this is b
            yes its b
            yes b

    I'm some more docs

    Returns:
        dict: the special thing
        list: other

    alksdfj
    ''')

    # renders the docstring (looks the same as the input)
    print(doc)

    # see the breakdown of the different sections/parameters.
    print(repr(doc))

    doc['args'].append(doc.Param.new(
        'c', 'list', 
        'this is c, that holds some values.\\nsome more text.'))
    
    assert str(doc) == '''This is my docstring description.

    Args:
        a (str): this is a
        b (str): this is b
            yes its b
            yes b
        c (list): this is c, that holds some values.
            some more text.

    I'm some more docs

    Returns:
        dict: the special thing
        list: other

    alksdfj
    '''
"""
from __future__ import annotations
from typing import List, Union, cast, overload, TypeVar, Generic
import re
# import copy
import inspect

# https://realpython.com/documenting-python-code/#docstring-types

BodyType = TypeVar('BodyType', bound='Block|str')


class BaseBlock(Generic[BodyType]):
    '''This represents a block of text. This will separate out any leading or trailing blank lines
    as well as factor out the common indentation. This lets you safely modify the content of the
    block while preserving the surrounding whitespace. It also can use an optional section title
    with it's own separate indentation (like with google docstring sections).

    .. code-block:: python

        block = Block('\n\n\n    blah\n\n\n\n')
        block.body = ['blorg', 'blagh']
        assert str(block) == '\n\n\n    blorg\n    blagh\n\n\n\n'
    '''
    leading: list[str]
    trailing: list[str]
    _body: list[BodyType]

    INDENT_WIDTH = 4
    def __init__(self, text, title=None, name=None, kind=None, cleandoc=False, end_newline=True, raw=False, indent=0):
        self._body = []
        if cleandoc:  # NOTE: this fails if Block is given a list
            text = inspect.cleandoc(text)
        lines = aslines(text, end_newline=end_newline)
        self.leading, self.body, self.trailing, min_indent = (
            ([],lines,[],None) if raw else separate_whitespace(lines))

        # get the title/body indent offset
        min_indent = min_indent or 0
        title_indent = len(title) - len(title.lstrip()) if title else min_indent
        self.min_indent = min(title_indent, min_indent) + (indent or 0)
        self.child_indent = min_indent - title_indent
        # store everything
        self.kind = kind
        self.title = title = title.lstrip() if title else None
        self.name = name or title

    def __repr__(self):
        '''A string representation of the block that shows the 
        division of sections.
        '''
        return border(self._format_body(), f'{self.__class__.__name__}(name={self.name})')

    def __str__(self):
        '''The fully formatted string.'''
        return self._format_body()

    def __bool__(self):
        '''Checks if the block has any non-whitespace content. 
        Analogous to ``bool('    '.strip())``  
        '''
        return bool(
            self.title or 
            any(l.strip() if isinstance(l, str) else l for l in self.body) or 
            any(l.strip() for l in self.leading + self.trailing))

    def __eq__(self, other):
        '''Checks if the string representations are the same. Whitespace does count.'''
        return str(self) == str(other)

    def __iter__(self):
        '''Iterate over lines in the body. Does not include whitespace or title.'''
        return iter(self.body)

    # def format(self, mode='s'):
    #     '''Format the block. Equivalent to ``str(block)``.'''
    #     return self._format_body(mode=mode)

    def _format_body(self, body=None, mode='s'):  # type: ignore
        body = self.body if body is None else aslines(body, end_newline=True)
        if mode == 'r':  # allow drawing boxes around children
            body = [repr(l) if isinstance(l, Block) else l for l in body]
        # indent the body and make sure it ends with a new line
        body = [indent(str(l), self.min_indent + self.child_indent) for l in body]
        title = [indent(self.title, self.min_indent)] if self.title else []
        # join everything together
        return ''.join(map(str, title + self.leading + body + self.trailing))

    # lets you set body with a string.
    @property
    def body(self): return self._body
    @body.setter
    def body(self, value): 
        self._body = self._prepare_body_lines(aslines(value, end_newline=True))

    def _prepare_body_lines(self, lines: list[str]) -> list[BodyType]:
        raise NotImplementedError

    def children(self, kind=..., name=..., include_self=False):
        '''Return children recursively matching some query.
        
        A block can contain other blocks, so this could be used to iterate over 
        all "Example" or "Arguments" blocks, for example.

        Arguments:
            kind (str): The block kind to match - e.g. ``'args'``, ``'returns'``
            name (str): The block name to match. This is the exact name of the section,
                So if you had a custom section ``My Examples:``, you'd enter ``'my examples'``.
                
        .. note::

            The difference between ``name`` and ``kind`` is that ``kind`` has recognized 
            sections with multiple variations and will normalize the name so a single string
            e.g. ``Arguments, Args => ARGS``. This makes it easier to search for all ARGS
            sections without having to check for each possible variation. ``name`` is there 
            when you want to search with the actual name of the section.
        '''
        if include_self and (kind == ... or self.kind == kind.upper()) and (name == ... or (self.name or '').lower() == name.lower()):
            yield self
        for b in self.body:
            if isinstance(b, BaseBlock):
                yield from b.children(kind, name, include_self=True)

    def first(self, kind=..., name=...):
        '''Return the first child matching a query. See ``children`` for arguments.'''
        return next(iter(self.children(kind, name)), None)

    @overload
    def __getitem__(self, k: str|int) -> BodyType: ...
    @overload
    def __getitem__(self, k: slice) -> list[BodyType]: ...
    
    def __getitem__(self, k: int|slice|str) -> BodyType|list[BodyType]:
        '''Return a direct child matching a name or index. Basically, you can 
        index like a list or a dict.
        '''
        if isinstance(k, (int, slice)):
            return self.body[k]
        for b in self.body:
            if nocaseeq(getattr(b, 'name', None), k):
                return b
        raise KeyError(k)

    def __delitem__(self, k: int|slice|str):
        '''Delete a line from the Block.'''
        if isinstance(k, (int, slice)):
            del self.body[k]
        else:
            for i, b in enumerate(self.body):
                if nocaseeq(getattr(b, 'name', None), k):
                    del self.body[i]
                    return
            raise KeyError(k)

    def get(self, k: str|int, default=None):
        try:
            return self[k]
        except (KeyError, IndexError):
            return default

    # basic manipulation

    def prepend(self, *x: BodyType):
        '''Prepend lines to the body.'''
        self.body[0:0] = x
        return self

    def append(self, *x: BodyType):
        '''Append lines to the body.'''
        self.body.extend(x)
        return self

    def indent(self, n: int=1, width: int|None=None):
        '''Indent the entire block.'''
        width = self.INDENT_WIDTH if width is None else width
        self.min_indent = max(self.min_indent + n * width, 0)
        return self

    def dedent(self, n: int=1, width: int|None=None):
        '''Dedent the entire block.'''
        return self.indent(-n, width)

    def set_indent(self, indent: int):
        if indent is not None:
            self.min_indent = indent
        return self

    # def strip(self): # XXX: THIS SHOULDNT HAPPEN INPLACE !!
    #     '''Strip whitespace from the block.'''
    #     self.leading, self.trailing = [], []
    #     return self

    def section_partition(self, pattern: str, get_kind=None, break_unnamed_sections=None):
        self.body = _section_partition(''.join(map(str, self.body)), pattern, get_kind, break_unnamed_sections)
        return self




class Block(BaseBlock[Union[str, BaseBlock[str]]]):
    def _prepare_body_lines(self, lines: list[str]) -> list[str]:
        return lines


class Param(Block):
    '''Represents the text belonging to a single parameter. 

    Most of this code is just to facilitate the parsing, 
    changing, and reformatting of the parameter data.
    '''
    _keys = ()
    _format=None
    pattern: re.Pattern|str = ''
    name = dtype = desc = None
    changed = False
    def __init__(self, text, can_be_unnamed=False, block_kind=None, **kw):
        self.__data = {}
        self.pattern = p = re.compile(self.pattern)
        self._keys = self._keys or [k for k, i in sorted(p.groupindex.items(), key=lambda x: x[1])]
        self._can_be_unnamed = can_be_unnamed
        self.block_kind = block_kind

        super().__init__(text + '\n' if not text.endswith('\n') else text, kind='PARAM', **kw)        

        # parse out the data
        m = re.match(p, ''.join(map(str, self.body)) + '\n')
        self.__data = self.prepare(**(m.groupdict() if m else {}))
        self.__dict__.update(self.__data)
        self.changed = False

    def __setattr__(self, k, v):
        if k in self._keys:  # track changes
            self.__data[k] = v
            self.changed = True
        object.__setattr__(self, k, v)

    def get(self, k):
        return getattr(self, k, None)

    def update(self, **kw):  # update multiple
        if set(kw) - set(self._keys):
            raise TypeError(set(kw) - set(self._keys))
        self.__data.update(kw)
        self.__dict__.update(kw, changed=True)

    def _format_body(self, body=None, *a, **kw):
        if self.changed and body is None:  # the param text was changed - recompile the text
            self.body = self.format(**self.__data).splitlines(keepends=True)
            self.changed = False
        return super()._format_body(body, *a, **kw)

    def format(self, **kw):
        raise NotImplementedError

    def replace(self, **kw):
        return self.__class__(
            self.format(**dict(self.__data, **kw)), 
            can_be_unnamed=self._can_be_unnamed)

    @classmethod
    def new(cls, *a, **kw):
        return cls(cls.format(cls, *a, **kw))

    def prepare(self, **kw):
        return kw


class Docstring(BaseBlock[Block]):
    '''This represents an entire docstring with each section '''
    HEADER_GROUPS = dict(
        ARGS=["Arguments", "Args", "Parameters", "Params"],
        EXCEPT=["Raises", "Exceptions", "Except"],
        ATTRS=["Attributes"],
        EXAMPLE=["Example", "Examples"],
        RETURN=["Returns"],
        YIELD=["Yields"])
    SUPPORTS_PARAMS = ('ARGS', 'ATTRS')
    SUPPORTS_UNNAMED_PARAMS = ('RETURN', 'YIELD', 'EXCEPT')
    FIRST_BLOCK_DEFAULT_NAME = 'Description'
    FIRST_BLOCK_DEFAULT_KIND = 'DESC'

    header_format = r'^({}):? *\n'
    class Param(Param): pass

    def __init__(self, doc=None, name=None, cleandoc=True):
        if callable(doc):
            name = name or doc.__name__
            doc = getattr(doc, '__doc__', None)
        self.doc = doc
        self.name = name

        super().__init__(doc or '', name=name, cleandoc=cleandoc, raw=True)
        self.parse()

    def __repr__(self):
        return border(self._format_body(mode='r'), f'{self.__class__.__name__}(name={self.name})')

    def parse(self):
        kinds = {v: k for k, vs in self.HEADER_GROUPS.items() for v in vs}

        self.body = body = _section_partition(
            ''.join(map(str, self.body)), 
            self.header_format.format("|".join(kinds)), 
            kinds.get, 
            self._break_unnamed_sections)
        for i, b in enumerate(body):
            self.handle_section(b, i)
        return self

    def _break_unnamed_sections(self, body):
        return [body]

    def handle_section(self, block: Block, i: int):
        if not i:
            block.name = block.name or self.FIRST_BLOCK_DEFAULT_NAME
            block.kind = block.kind or self.FIRST_BLOCK_DEFAULT_KIND
        if block.kind in self.SUPPORTS_UNNAMED_PARAMS:
            block.body = [self.Param(l, can_be_unnamed=True) for l in _group_indents(block.body)]
        elif block.kind in self.SUPPORTS_PARAMS:
            block.body = [self.Param(l) for l in _group_indents(block.body)]
        return block

    def children(self, *a, **kw):
        for p in self.body:
            yield from p.children(*a, include_self=True, **kw)

    def first(self, *a, **kw):
        return next(iter(self.children(*a, **kw)), None)

    def _prepare_body_lines(self, lines: list[str]) -> list[Block]:
        return [Block(l) if not isinstance(l, Block) else l for l in lines]



class Google(Docstring):
    '''Google docstring parser.

    .. code-block:: python

        ds = Google("""This is my docstring description.

        Args:
            a (str): this is a



            b (str): this is b
                yes its b
                yes b

        I'm some more docs

        Returns:
            dict: the special thing
            list: other

        alksdfj

        .. code-block:: python

            print("this is something")

            print("hmm nice")


        ksnkdsjkdsksd ljksdfkjl
        asjkldf

        .. code-block:: python

            print("neat!")
        """)
    '''
    header_format = r'^({}):? *\n'

    def _break_unnamed_sections(self, body):
        '''This breaks of mid doc text (in between sections)'''
        return _break_minimum_indent(body)

    class Param(Param):
        '''The Google parameter format.

        .. code-block::

            {name} ({type}): {description}
        
        '''
        pattern = r'^ *(?P<name>\w*) *(?:\((?P<dtype>[^)]+)\))?: *(?P<desc>(?:\n|.)*)'
        def format(self, name=None, dtype=None, desc=None):
            return '{}: {}'.format(
                f'{name} ({dtype})' if name and dtype else name or dtype,
                # ' '.join(filter(None, [name, f'({dtype})' if dtype else None])), 
                indent(desc).lstrip(' '))

        def prepare(self, name=None, dtype=None, desc=None):
            # if self.block_kind != 'ARGS' and not dtype:
            #     name, dtype = None, name
            return dict(name=name, dtype=dtype, desc=inspect.cleandoc(desc or ''))


class Numpy(Docstring):
    '''Numpy docstring parser.

    .. code-block:: python

        ds = Numpy("""Gets and prints the spreadsheet's header columns

        Parameters
        ----------
        file_loc : str
            The file location of the spreadsheet
        print_cols : bool, optional
            A flag used to print the columns to the console (default is False)
            asjdfklasjdfkl

        Returns
        -------
        list
            a list of strings representing the header columns
        """)
    '''
    header_format = r'^({}) *\n[-=]+ *\n'

    def _break_unnamed_sections(self, body):
        return _break_full_newline(body)

    class Param(Param):
        '''The Numpy parameter format.

        .. code-block::

            {name} : {type}
                {description}
        
        '''
        pattern = r'^ *(?P<name>\w+) *(?:: *(?P<dtype>[^\n]+))? *\n(?P<desc>(?: +.*\n)*)'
        def format(self, name=None, dtype=None, desc=None):
            return '{}\n{}'.format(' : '.join(filter(None, [name, dtype])), indent(desc))


# block utils

def _break_sections(doc, pattern):
    '''Break text and body pairs for doc sections with headers.'''
    # find all matches
    matches = list(re.finditer(pattern, doc, flags=re.M))

    yield '', doc[:matches[0].start() if matches else None]
    for m1, m2 in zip(matches, matches[1:] + [None]):  # type: ignore
        # select the title of the heading
        title = doc[m1.start():m1.end()]
        body = doc[m1.end():m2 and m2.start()]
        yield title, body

def _section_partition(doc, pattern, get_kind=None, break_unnamed_sections=None):
    sections = []
    for title, body in _break_sections(doc, pattern):
        # pull info from title
        match = re.match(pattern, title) if title else None
        name = match.group(1) if match else None
        kind = get_kind(name) if get_kind else None
        body, *others = break_unnamed_sections(body) if break_unnamed_sections else [body]

        sections.append(Block(body, title, name, kind=kind))
        for other in others:
            sections.append(Block(other))
    return sections


def _break_minimum_indent(section):
    '''Break indent when the indentation drops below the indentation of the first line.
    
    Example

    .. code-block:: rst

            arg1 (int): laksdfjklaj

        --- break here ---
        This is some discussion
    '''
    lines = section.splitlines(keepends=True)
    indent = next((
        len(l) - len(l.lstrip()) for l in lines 
        if l.strip()), None)
    if indent is None:
        return [lines]

    i_break = next((
        i for i, l in enumerate(lines) 
        if l.strip() and len(l) - len(l.lstrip()) < indent
    ), None)
    if i_break is None or i_break >= len(lines):
        return [lines]
    return [lines[:i_break], lines[i_break:]]


# def _break_full_newline(section):
#     '''Break after an entirely blank line'''
#     lines = section.splitlines(keepends=True)
#     i_start = next((i+1 for i, l in enumerate(lines) if l.strip()), None)
#     i_break = next((i+1 for i, l in enumerate(lines) if i >= i_start and not l.strip()), None)
#     if i_break is None or i_break >= len(lines):
#         return [lines]
#     return [lines[:i_break], lines[i_break:]]

def _ibreak_full_newline(section):
    '''Break after an entirely blank line'''
    lines = section.splitlines(keepends=True)
    i_blank = -1
    offset = 0
    for i, l in enumerate(lines):
        i_blank = i if i_blank < offset and not l.strip() else i_blank
        if i_blank > offset and l.strip():
            yield lines[offset:i]
            offset = i
    if not offset or offset < len(lines):
        yield lines[offset:]

def _break_full_newline(section):
    return list(_ibreak_full_newline(section))



def _group_indents(lines):
    '''Break into indentation groups (one for every top level group). e.g. google arguments.

    Example

    .. code-block:: rst

            arg1 (int): laksdfjklaj
                jasflkjdslkf
            --- break here ---
            arg2 (int): lkasdfjlaksdf
            --- break here ---
            arg3 (int): laksfdkjla
    '''
    indents = [len(l) - len(l.lstrip()) for l in lines]
    min_indent = min((i for i, l in zip(indents, lines) if l.strip()), default=0)
    tops = [i for i, (idt, l) in enumerate(zip(indents, lines)) if l.strip() and min_indent >= idt]
    groups = []
    if tops and tops[0]:
        groups.append(lines[:tops[0]])
    for i, j in zip([0] + tops, tops + [None]):  # type: ignore
        if i != j:
            groups.append(lines[i:j])
    return [''.join(ls) for ls in groups]


# text utils

def indent(text, n=4):
    '''Indent a block of text'''
    return ''.join([' '*n + l for l in str(text).splitlines(keepends=True)])

def comment(text, ch='#'):
    '''Add a prefix to each line of a text block. Like commenting out code.'''
    return ''.join(f'{ch} {l}' for l in str(text).splitlines(keepends=True))

def border(text, title=None, top_ch='-', side_ch='|', n_top=3):
    '''Adds a border to the left and top sides, with an optional title.'''
    return top_ch * n_top + (f' {title} {top_ch}' if title else '') + '\n' + comment(text, side_ch)

def nocaseeq(a, b):
    '''Compare two values case-insensitively.'''
    return (a.lower() if isinstance(a, str) else a) == (b.lower() if isinstance(b, str) else b)

def aslines(text, end_newline=True) -> list[str]:
    '''Convert a block of text to lines, preserving new lines.'''
    lines = text.splitlines(keepends=True) if isinstance(text, str) else text or []
    if end_newline and lines and isinstance(lines[-1], str) and not lines[-1].endswith('\n'):
        lines[-1] = lines[-1] + '\n'
    return lines  # type: ignore


# def separate_whitespace(lines):
#     '''Separate whitespace above, below, and common indentation for a block of text.'''
#     # break off the leading and trailing spaces
#     nonblank = [i for i, l in enumerate(lines) if l.strip()]
#     i, j = (nonblank[0], nonblank[-1]+1) if nonblank else (0,0)
#     leading, body, trailing = lines[:i], lines[i:j], lines[j:]

#     # separate off the indentation from the body / title
#     min_indent = min((
#         len(l) - len(l.lstrip()) for l in body 
#         if l.strip()), default=None)
#     body = [l[min_indent:] if l.strip() else l for l in body]
#     return leading, body, trailing, min_indent


def separate_whitespace(lines):
    '''Separate whitespace above, below, and common indentation for a block of text.'''
    # break off the leading and trailing spaces
    is_block = [isinstance(l, Block) for l in lines]
    has_content = [is_block[i] or bool(l.strip()) for i, l in enumerate(lines)]
    nonblank = [i for i, l in enumerate(lines) if has_content[i]]
    i, j = (nonblank[0], nonblank[-1]+1) if nonblank else (0,0)
    leading, body, trailing = lines[:i], lines[i:j], lines[j:]
    is_block, has_content = is_block[i:j], has_content[i:j]
    # print(leading, body, trailing)

    # separate off the indentation from the body / title
    min_indent = min((
        l.min_indent if is_block[i] else len(l) - len(l.lstrip()) 
        for i, l in enumerate(body) if has_content[i]), default=None)
    body = [
        l.set_indent(min_indent) if is_block[i] else 
        l[min_indent:] if has_content[i] else l
        for i, l in enumerate(body)
    ]
    return leading, body, trailing, min_indent


STYLES = {
    'google': Google,
    'numpy': Numpy,
}

def parse(func, style: str|None=None, **kw):
    for s in [style] if isinstance(style, str) else style or list(STYLES):
        return STYLES[s](func, **kw).parse()


# def _resub_groups(pattern, body, *a, **kw):  # FIXME: how to handle missing ???
#     '''Replace the content of regex capture groups.'''
#     p = re.compile(pattern, re.M) if not isinstance(pattern, re.Pattern) else pattern
#     names, idxs = map(list, zip(*sorted(p.groupindex.items(), key=lambda x: x[1])))
#     a = list(a) + [kw.pop(n, None) for n in names[len(a):]]

#     m = p.match(body)
#     joined = ''
#     for k, kj, x in zip(idxs, idxs[1:]+[None], a):
#         print('replacing', k and m.end(k), kj and m.start(kj), repr(str(x)), repr(body[k and m.end(k):kj and m.start(kj)]))
#         joined += f'{x}' + body[k and m.end(k):kj and m.start(kj)]
#     print('substituted:', repr(joined))
#     return joined


if __name__ == '__main__':

    ds = Google('''This is my docstring description.

    Args:
        a (str): this is a



        b (str): this is b
            yes its b
            yes b

    I'm some more docs

    Returns:
        dict: the special thing
        list: other

    alksdfj

    .. code-block:: python

        print("this is something")

        print("hmm nice")


    ksnkdsjkdsksd ljksdfkjl
    asjkldf

    .. code-block:: python

        print("neat!")
    ''').parse()

    ds['args'].append(ds.Param.new('somethingggg', None, 'kjlsadjkfdsajkl\ndfgad\nasdf'))
    ds['args'].append(ds.Param.new(None, 'dict', '819723789'))
    ds['args'].append(ds.Param.new(
        'nested_kw', 'dict', Block([
            # '\n', '\n',
            'nested parameters that go to somewhere else\n',
            ds.Param.new('a', 'int', 'aaa'),
            ds.Param.new('b', 'int', 'aaa'),
        ])))

    print(repr(ds))
    print(ds)

    ds2 = Google('''This is my docstring description.

    Args:
        x (str): this is x
        y: im y
          yes hi

    alksdfj
    ''').parse()

    ds['args'].append(*ds2['args'].body)

    ds['args']['somethingggg'].name = 'bsak'

    print(repr(ds))
    print(ds)

    # print(ds.compose())

    ds = Numpy('''Gets and prints the spreadsheet's header columns

    Parameters
    ----------
    file_loc : str
        The file location of the spreadsheet
    print_cols : bool, optional
        A flag used to print the columns to the console (default is False)
        asjdfklasjdfkl

    Returns
    -------
    list
        a list of strings representing the header columns
    ''').parse()

    ds['Parameters'].append(ds.Param.new('somethingggg', None, 'kjlsadjkfdsajkl\n  dfgad\n    asdf'))
    ds['Parameters'].append(ds.Param.new(None, 'dict', '819723789'))

    # print(ds.compose())
    print(repr(ds))
    ds['Parameters']['somethingggg'].name = 'bsak'
    p = ds['Parameters']['dict']
    p.name = 'bloop'
    p.desc = 'hello\nasdf'
    print(repr(p))
    print(ds)

    # print(ds.compose())



'''Create an argparse ArgumentParser from a function signature.

.. note::

    This isn't a fully featured CLI alternative. This is meant for cases
    where you need to interface with another CLI that uses argparse 
    (e.g. pytorch_lightning).

I find writing out argparse arguments extremely annoying. 
Why do I have to? The information is already there!
Usually I use ``fire``, but that's not always an option so 
hopefully this can help reduce the amount of code you need
to write.


.. code-block:: python

    import starstar.argparse

    def myfunc(batch_size=64, dataset='urbansas', n_mels=128, win_size=10):
        ...

    parser = starstar.argparse.from_func(myfunc)
    args = parser.parse_args()

    # this.py --batch-size 32 --win-size 15



'''
from __future__ import annotations
import copy
import functools
import os
from typing import Any, Callable, Union, get_type_hints
try:
    from typing import Literal, get_origin, get_args
except ImportError:
    from typing_extensions import Literal, get_origin, get_args
import argparse
import inspect
import starstar
from starstar import docstr, POS_ONLY, VAR_POS, VAR_KW, POS_KW
from starstar.parse import parse

TYPE = type

# improved argparser with vararg parsing

class ArgumentParser(argparse.ArgumentParser):
    accept_var_kw = False
    def __init__(self, *a, **kw):
        # allow newlines in arguments
        kw.setdefault('formatter_class', argparse.RawDescriptionHelpFormatter)
        kw.setdefault('conflict_handler', 'resolve')
        super().__init__(*a, **kw)

    def parse_args(self, *a):
        if getattr(self, 'accept_var_kw', False):
            return self.parse_varkw_args(*a)
        return super().parse_args(*a)

    def parse_varkw_args(self, *a):
        '''Parse arguments and accept var kwargs if provided.'''
        # support variable kwargs
        parsed, unknown = self.parse_known_args(*a)
        key = None
        values = []
        self2 = copy.copy(self)
        for arg in unknown + [None]:  # type: ignore
            if arg is None or arg.startswith(("-", "--")):
                if key is not None:
                    self2.add_argument(key, required=False, type=parse, default=argparse.SUPPRESS)
                key = arg
                values.clear()
                continue
            values.append(arg)
        return super(ArgumentParser, self2).parse_args(*a)


def from_any(
        obj: Callable|dict, #|list|tuple
        *,
        parser: argparse.ArgumentParser|None=None, 
        subparsers: argparse.ArgumentParser|None=None,
        parser_name: str|None=None,
        _cmd_prefix: str='',
        **kw
):
    if callable(obj) or isinstance(obj, type):
        return from_func(obj, parser=parser, subparsers=subparsers, parser_name=parser_name, **kw)

    # create parser if not provided
    if parser is None:
        if subparsers:
            assert parser_name
            parser = subparsers.add_parser(parser_name, help='asdf')
        else:
            parser = ArgumentParser(**kw)
    # subparsers = parser._subparsers
    # if subparsers is None:
    subparsers = parser.add_subparsers(dest=f'__{_cmd_prefix or parser_name or ""}command', help='Available commands.')
    # FIXME: how to handle nested commands?
    
    # handle different object types
    if isinstance(obj, (list, tuple)):
        assert all(callable(o) or isinstance(o, type) for o in obj)
        obj = {func.__name__: func for func in obj}

    obj_for_parser = obj
    if isinstance(obj, dict):
        for name, obj_i in obj.items():
            from_any(obj_i, subparsers=subparsers, parser_name=name, _cmd_prefix=f'{_cmd_prefix or ""}__{name}')
    else:
        # parser = from_any(vars(obj), **kw)
        raise TypeError("Object must be a function or a dict of functions.")
    parser._calling_object = obj_for_parser
    return parser


def from_func(
        func: Callable, 
        *, 
        parser: argparse.ArgumentParser|None=None, 
        subparsers: argparse.ArgumentParser|None=None, 
        parser_name: str|None=None, 
        docstyle: str|None=None, 
        prog: str|bool|None=None, 
        actions: dict[str, str]|None=None, 
        choices: dict[str, list]|None=None, 
        defaults: dict[str, Any]|None=None,
        consts: dict[str, Any]|None=None, 
        nargs: dict[str, int|str]|None=None, 
        types: dict[str, Callable]|None=None, 
        dests: dict[str, str]|None=None, 
        # short: dict[str, str]|None=None,
        env_format: str|None=None,
        **kw
):
    '''Create an argparse ArgumentParser from a function!
    
    Arguments:
        func (callable):
        parser (argparse.ArgumentParser): You can pass in your own parser to use.
        docstyle (str): The name of the docstring style. Can be google or numpy.

    '''
    actions = actions or {}
    choices = choices or {}
    defaults = defaults or {}
    consts = consts or {}
    nargs = nargs or {}
    dests = dests or {}
    types = types or {}

    # 
    # if 'parents' in kw:
    #     kw.setdefault('add_help', False)

    # handle class init
    cls = None
    if isinstance(func, type):
        cls = func
        func_name = func.__name__
        docs = func.__init__.__doc__ or func.__doc__
        func = func.__init__
    else:
        func_name = func.__name__
        docs = func.__doc__


    # parse docstring
    docs = docstr.parse(docs, style=docstyle) or None
    docargs = list(docs.children('args')) if docs else None
    description = docs.first('desc') if docs else None
    desc_str = str(description) if description else None

    # create parser if not provided
    if not parser:
        if subparsers:
            help_msg = desc_str.split('\n')[0] if desc_str else ''
            parser = subparsers.add_parser(parser_name, help=help_msg or 'hi')
        else:
            parser = ArgumentParser(**kw)
    parser._calling_object = cls if cls is not None else func
    # use the function docstring description
    if desc_str and 'description' not in kw:
        parser.description = desc_str
    # prog is the "name of the program" as displayed by the cli
    # by default it is the script name e.g. myscript.py
    prog = func_name if prog is True else prog
    if prog:
        parser.prog = prog

    # get signature
    s = starstar.signature(func)
    # short = _get_shortkw(s.parameters, reserved=['h'], short=short)
    # short = {}

    # fix union syntax and get type hints
    s = copy.copy(s)
    anns = dict(func.__annotations__)
    for n, tstr in list(anns.items()):
        anns[n] = _repl_union(tstr)
    type_hints = get_type_hints(TYPE(func_name, (object,), {'__annotations__': anns}))

    # make any arguments before varargs positional only
    override_pos_only = any(p.kind == VAR_POS for p in s.parameters.values())

    # create each parameter
    for name, p in s.parameters.items():
        pkw = {}

        flag_name = name
        # short_name = short.get(name)

        # check if the argument is positional
        var_pos = p.kind == VAR_POS
        pos_only = p.kind == POS_ONLY
        pos_kw = p.kind == POS_KW
        positional = var_pos or pos_only or (override_pos_only and pos_kw)

        # var arg specifics: *a, **kw
        if var_pos:
            # allow 0 or more
            pkw['nargs'] =  '*'
            # indicate that it gobbles up positional args
            pkw['metavar'] = f'*{name}'
            override_pos_only = False
        elif p.kind == VAR_KW:
            # can't handle varkw in here (see parse_args)
            parser.accept_var_kw = True
            continue

        # manual nargs
        if name in nargs:
            pkw['nargs'] = nargs[name]
        # manual action
        if name in actions:
            pkw['action'] = actions[name]
        # custom type
        if name in types:
            pkw['type'] = types[name]
        # rename destination
        if name in dests:
            pkw['dest'] = dests[name]
        
        # set defaults
        default = defaults.get(name, p.default)
        if default == inspect._empty:
            if not positional:
                pkw['required'] = True
        else:
            if positional and 'nargs' not in pkw:
                pkw['nargs'] = '?'
            pkw['default'] = default
        if env_format:
            env_key = env_format.format(name.upper())
            if env_key in os.environ:
                pkw['default'] = os.environ[env_key]

        # set action & type

        # use type annotation
        dtype = type_hints.get(name)
        if dtype:
            if _lenient_subclass(dtype, bool):
                if 'action' not in pkw:
                    if default is True:
                        pkw['action'] = 'store_false'
                        flag_name = f'no-{flag_name}'
                        # if short_name:
                        #     short_name = f'no{short_name}'
                    else:
                        pkw['action'] = 'store_true'
            if _lenient_subclass(dtype, (list, tuple)):
                if 'action' not in pkw:
                    pkw['action'] = 'extend'
                if 'nargs' not in pkw:
                    pkw['nargs'] = '+'
            elif 'type' not in pkw:
                pkw['type'] = _parse_type_checker(dtype)
        if 'type' not in pkw:
            pkw['type'] = parse

        # manage constants
        if name in consts:
            if pkw.get('action') in {'append', 'extend'}:
                pkw['action'] = 'append_const'
            elif 'action' not in pkw:
                pkw['action'] = 'store_const'
            pkw['const'] = consts[name]

        if pkw.get('action') in ('append_const', 'store_const', 'store_true', 'store_false'):
            pkw.pop('type', None)

        # set choices
        if name in choices:
            pkw['choices'] = choices[name]
        elif type and get_origin(type) == Literal:
            pkw['choices'] = get_args(type)

        # create flags
        flag_name = flag_name.replace('_', '-')
        if positional:
            argnames = [flag_name]
        else:
            argnames = [f'-{flag_name}', f'--{flag_name}']

        # find the help message
        for darg in docargs or []:
            for dp in darg.body:
                if dp.name == name:
                    helpmsg = dp.get('desc')
                    if helpmsg:
                        pkw['help'] = str(helpmsg)
                        break

        parser.add_argument(*argnames, **pkw)
    return parser


# # short arguments

# def _get_shortkw(parameters, ignore=(), reserved=(), short=None):
#     '''Given a set of arguments, find the natural short arguments e.g. ``--wow -> -w``'''
#     short = short or {}
#     for k in parameters:
#         if parameters[k].kind == POS_ONLY:
#             continue
#         if k not in ignore and k[0] not in reserved and k[0] not in short:
#             short[k] = k[0]
#     return short



# type checkers

def _lenient_subclass(obj, cls, union_cond=all):
    '''Check isinstance/subclass including unions.'''
    if get_origin(obj) == Union:
        return union_cond(t is None or t is type(None) or _lenient_subclass(t, cls) for t in get_args(obj))
    return isinstance(obj, cls) or isinstance(obj, type) and issubclass(obj, cls)

def _type_check(type, x):
    '''Cast union type'''
    if get_origin(type) == Union:
        for t in get_args(type):
            try:
                return _type_check(t, x)
            except Exception:
                pass
        raise TypeError(f"{x} could not be cast to {type}")
    return type(x)

def _parse_type_checker(type):
    @functools.wraps(type)
    def __type(x): return _type_check(type, parse(x))
    if __type.__name__ == '__type':
        __type.__name__ = __type.__qualname__ = _type_check_name(type)
    return __type

# def _type_checker(type):
#     '''A decorator for type checking.'''
#     @functools.wraps(type)
#     def __type(x):  return _type_check(type, x)
#     if __type.__name__ == '__type':
#         __type.__name__ = __type.__qualname__ = _type_check_name(type)
#     return __type

def _type_check_name(type):
    '''Get the name of a type check (including unions)'''
    if get_origin(type) == Union:
        return '|'.join([_type_check_name(t) for t in get_args(type)])
    return type.__name__



# Fix old union annotations

# Constants
OPEN_BRACKET = '['
CLOSE_BRACKET = ']'
COMMA = ','
OR = '|'


def _repl_union(s: str):
    """ Replace PEP 604-style annotations (i.e. like `X | Y`) with `Union[X, Y]`."""
    # If there is no '|' character in the annotation part, we just return it.
    if OR not in s:
        return s
    s = s.replace(' ', '')

    # Checking for brackets like `List[int | str]`.
    if OPEN_BRACKET in s:
        # Get any indices of COMMA or OR outside a braced expression.
        outer_commas, outer_pipes = _outer_comma_and_pipe_indices(s)
        # commas outside bracket? e.g. dict[str | int, str]***,*** value[test]
        if outer_commas:
            return COMMA.join([_repl_union(i) for i in _sub_strings(s, outer_commas)])
        # pipes outside bracket? e.g.: value | dict[str | int, list[int | str]]
        if outer_pipes:
            return f'Union{OPEN_BRACKET}{COMMA.join([_repl_union(i) for i in _sub_strings(s, outer_pipes)])}{CLOSE_BRACKET}'

        # no outer commas/pipes, and `SomeType[str][bool]` is  invalid
        # Replace inside brackets, e.g.: dict[str | int, str]
        first_start_bracket = s.index(OPEN_BRACKET)
        last_end_bracket = s.rindex(CLOSE_BRACKET)
        return (
            f'{s[:first_start_bracket]}{OPEN_BRACKET}'
            f'{_repl_union(s[first_start_bracket + 1:last_end_bracket])}'
            f'{CLOSE_BRACKET}{s[last_end_bracket + 1:]}')

    elif COMMA in s:  #  e.g. `int | str, float | None`
        return COMMA.join([_repl_union(i) for i in s.split(COMMA)])

    return f'Union{OPEN_BRACKET}{s.replace(OR, COMMA)}{CLOSE_BRACKET}'  # e.g. `int | str`


def _sub_strings(s: str, split_indices):
    """Split a string on the specified indices, and return the split parts."""
    prev = -1
    for idx in split_indices:
        yield s[prev+1:idx]
        prev = idx
    yield s[prev+1:]


def _outer_comma_and_pipe_indices(s: str):
    """Return any indices of ',' and '|' that are outside of braces."""
    indices = {OR: [], COMMA: []}
    brace_dict = {OPEN_BRACKET: 1, CLOSE_BRACKET: -1}
    brace_count = 0

    for i, char in enumerate(s):
        if char in brace_dict:
            brace_count += brace_dict[char]
        elif not brace_count and char in indices:
            indices[char].append(i)
    return indices[COMMA], indices[OR]



# calling the function


def call_any(parser, args, _cmd_prefix=''):
    # pull out the command from the arguments (so we don't pass it to the function)
    a = dict(args)
    cmd = a.pop(f'__{_cmd_prefix or ""}command', None)

    # it's a nested object, keep going
    if cmd is not None:
        # from IPython import embed
        # embed()
        return call_any(next(
            a._name_parser_map[cmd]
            for a in parser._subparsers._group_actions
            if cmd in a._name_parser_map
        ), a, _cmd_prefix=f'{_cmd_prefix}__{cmd}')
    
    # no more sub-commands
    obj = parser._calling_object
    if not callable(obj):
        parser.print_help()
        parser.exit(1)

    # call the function
    a, kw = starstar.as_args_kwargs(obj, a)
    return obj(*a, **kw)



# Demo


def Star(func, **kw):
    parser = from_any(func, **kw)
    # from IPython import embed
    # embed()
    args = parser.parse_args()
    call_any(parser, vars(args))


if __name__ == '__main__':
    def myfunction(aaa: int, bbb=5, *a, items: list, indices: int|list, wow: bool=False, quoi='aaa', **kw):
            '''Look at my function

            I love it so much

            Arguments:
                aaa: first
                bbbb: second thing
                wow (int): ok cool
                quoi (str): wow alright
            '''
            print(aaa,  bbb, a, items, indices, wow, quoi, kw)


    def my_other_thing(a, b, c):
        print(a, b, c)

    Star({
        'xxx': myfunction,
        'yyy': my_other_thing,
        'aaa': {
            'abc': myfunction
        },
    })
